import logging.config
from functools import reduce


__all__ = 'configure', 'resolve'


class Configurator(logging.config.BaseConfigurator):

  def __init__(self, config):
    self.config = self.create_dict(config)

  def create_dict(self, config):
    d = ConvertingDict(config)
    d.configurator = self
    d.convert_items()
    return d

  def convert(self, value):
    if isinstance(value, dict):
      if not isinstance(value, ConvertingDict):
        value = self.create_dict(value)
      if '()' in value:
        value = self.configure_custom(value)

    return super(Configurator, self).convert(value)


class ConvertingDict(dict, logging.config.ConvertingMixin):

  def __init__(self, *args, **kwargs):
    dict.__init__(self, *args, **kwargs)

  def convert_items(self):
    for k, v in self.items():
      self.convert_with_key(k, v)


def merge(target, source):
  '''Deep ``dict`` merge'''

  result = target.copy()  # shallow copy
  stack  = [(result, source)]
  while stack:
    currentTarget, currentSource = stack.pop()
    for key in currentSource:
      if key not in currentTarget:
        currentTarget[key] = currentSource[key]  # appending
      else:
        if isinstance(currentTarget[key], dict) and isinstance(currentSource[key], dict):
          currentTarget[key] = currentTarget[key].copy()  # nested dict copy
          stack.append((currentTarget[key], currentSource[key]))
        elif currentTarget[key] is not None and currentSource[key] is None:
          del currentTarget[key]  # remove key marked as None
        else:
          currentTarget[key] = currentSource[key]  # overriding

  return result


def configure(*args):
  '''Merged nested configuration dictionary is wrapped into a resolver'''

  merged = reduce(merge, args)
  return Configurator(merged).config


resolve = logging.config._resolve




import unittest

from . import configure


class TestEnvconf(unittest.TestCase):

  dev = None
  prd = None


  def setUp(self):
    base = {
      'global' : {
        'server.socket_host' : '127.0.0.1',
        'server.socket_port' : 8080,
        'server.thread_pool' : 8,
      },
      'logging' : {
        'version'    : 1,
        'formatters' : {
          'print' : {
            'format' : '%(asctime)s %(levelname)-8s %(name)-15s %(message)s'
          },
        },
        'handlers' : {
          'console' : {
            'level'     : 'DEBUG',
            'class'     : 'logging.StreamHandler',
            'formatter' : 'print'
          },
        },
        'loggers' : {
          'app' : {
            'handlers'  : ['console'],
            'level'     : 'INFO',
            'propagate' : False
          },
        }
      },
      'app' : {
        'api' : {
          '/' : {
            'request.dispatch' : {
              '()'       : 'xml.parsers.expat.ParserCreate',
              'encoding' : 'utf-8'
            }
          }
        },
      }
    }

    production = (base, {
      'global' : {
        'server.socket_host' : '0.0.0.0',
        'server.thread_pool' : 16,
      },
      '/' : {
        # compression
        'tools.gzip.on'         : True,
        'tools.gzip.mime_types' : ['application/json', 'application/javascript'],
        # authentication
        'tools.auth_basic.on'            : True,
        'tools.auth_basic.realm'         : 'App',
        'tools.auth_basic.checkpassword' : 'ext://clor.test.TestEnvconf'
      }
    })

    development = (base, {
      'global' : {
        'server.thread_pool' : None,
      },
      'logging' : {
        'loggers' : {
          'app' : {
            'level' : 'DEBUG'
          }
        }
      },
      'app' : {
        'api' : {
          '/' : {
            'tools.response_headers.on'      : True,
            'tools.response_headers.headers' : [('Access-Control-Allow-Origin', '*')]
          }
        }
      }
    })

    self.dev = configure(*development)
    self.prd = configure(*production)

  def testInherit(self):
    self.assertEqual('127.0.0.1', self.dev['global']['server.socket_host'])
    self.assertEqual(['console'], self.dev['logging']['loggers']['app']['handlers'])

  def testSet(self):
    self.assertEqual(True, self.prd['/']['tools.gzip.on'])

  def testOverride(self):
    self.assertEqual('0.0.0.0', self.prd['global']['server.socket_host'])
    self.assertEqual('DEBUG', self.dev['logging']['loggers']['app']['level'])

  def testRemove(self):
    self.assertEqual({
      'server.socket_host': '127.0.0.1',
      'server.socket_port': 8080
    }, self.dev['global'])

  def testExternal(self):
    self.assertIs(self.__class__, self.prd['/']['tools.auth_basic.checkpassword'])

  def testFactory(self):
    obj = self.dev['app']['api']['/']['request.dispatch']
    self.assertTrue(repr(obj).startswith('<pyexpat.xmlparser object'))

  def testFactoryItems(self):
    for _, v in self.dev['app']['api'].items():
      obj = v['request.dispatch']
      self.assertTrue(repr(obj).startswith('<pyexpat.xmlparser object'))

  def testUpdate(self):
    target = {}
    target.update(self.prd['/'])
    self.assertIs(self.__class__, target['tools.auth_basic.checkpassword'])




try:
  from setuptools import setup
except ImportError:
  from distutils.core import setup


setup(
  name             = 'Clor',
  version          = '0.2.0',
  author           = 'saaj',
  author_email     = 'mail@saaj.me',
  packages         = ['clor'],
  test_suite       = 'clor.test',
  url              = 'https://bitbucket.org/saaj/clor',
  license          = 'LGPL-2.1+',
  description      = 'Application configuration via Python logging.config',
  long_description = open('README.txt', 'rb').read().decode('utf-8'),
  platforms        = ['Any'],
  keywords         = 'python configuration-management',
  classifiers      = [
    'Topic :: Software Development :: Libraries',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: Implementation :: CPython',
    'Intended Audience :: Developers'
  ]
)

